local M = {}
M.type = "auxiliary"
M.relevantDevice = "engine"

local abs = math.abs
local min = math.min
local max = math.max
local floor = math.floor
local sqrt = math.sqrt
local rpmToAV = 0.104719755
local kphToMs = 1 / 3.6

local engine = nil
local iDM_Transmission = nil
local p1_Motors = nil
local p3_Motors = nil
local motors = nil
local throttleSet = nil
local batteryLow = 0.15
local batteryCriticalLow = 0.05
local engineTempHigh = 100
local engineTempReady = 90
local engineTempLow = 87
local freeRPM = 5500
local idleRPM = 1500
local parallelRPM = 1200
local freeAV = freeRPM * rpmToAV
local idleAV = idleRPM * rpmToAV
local parallelAV = parallelRPM * rpmToAV
local idleSpeed = 15 * kphToMs
local idlingTime = 0
local idlingBreak = 0
local lastBatteryLevel = 0
local redlineRPM = 5200
local RPMsyncing = false -- indicating whether is the rpm syncronizing in order to change from series to parallel

-----------------------------------------
--- for aggression calculation
local throttleAggression = 0
local holdTime = 0.5  -- time to hold throttleAggression at 1 after flooring
local holdCounter = 0
local increaseFactor = 3  -- adjustable increase factor for fast response
local decayFactor = 0.8 -- adjustable decay factor for slow decay
local minDecayFactor = 0.1 -- minimum boundary for decay speed
----------------------------------------
-----------------------------------
-- setting velocity threshold threshold for parallel-series conversion
-- (series_to_parallel must be GREATER than parallel_to_series)
local series_to_parallel = 1.2
local parallel_to_series = 0.95
-----------------------------------
-- low battery threshold for ev to force change to hev mode (min = 0, max = 1)
local ev_low_battery_exit = 0.15
-----------------------------------
-- high throttle aggression threshold for ev to force change to hev mode (min = 0, max = 1)
local ev_high_throttle_exit = 0.99
-----------------------------------
-- critically low battery threshold for force change to series (min = 0, max = 1)
-- (battery_critical_low must be LESS than ev_low_battery_exit)
local battery_critical_low = 0.05
-----------------------------------
local parallel_to_ev_threshold_factor = 0
local series_to_ev_threshold_factor = 0.3
local parallel_to_series_threshold_factor = 2
local series_to_parallel_threshold_factor = 1
-----------------------------------
local max_engine_stop_speed = 130 -- maximum speed that will allow engine stop in parallel mode
--value correctness check
if series_to_parallel < parallel_to_series then
  error("series_to_parallel must be GREATER than parallel_to_series, please set correct values")
end
if battery_critical_low > ev_low_battery_exit then
  error("battery_critical_low must be LESS than ev_low_battery_exit, please set correct values")
end
-----------------------------------


local unsmoothedCurrentPower = 0
local currentPowerSmoother = newExponentialSmoothing(100)
local engineCurrentPowerSmoother = newExponentialSmoothing(100)
local currentPowerCoefSmoother = newExponentialSmoothing(10000)

local function setMode(mode)
  electrics.values.hevState = mode
  if iDM_Transmission.mode ~= mode then
    currentPowerCoefSmoother:reset()
    iDM_Transmission:setMode(mode)
  end
end

local function updateMotorIgnition()
  if electrics.values.ignitionLevel >= 2 then
    for _, motor in ipairs(p1_Motors) do
      motor.ignitionCoef = 1
    end
    for _, motor in ipairs(p3_Motors) do
      motor.ignitionCoef = 1
    end
    for _, motor in ipairs(motors) do
      motor.ignitionCoef = 1
    end
  elseif electrics.values.ignitionLevel < 2 then
    for _, motor in ipairs(p1_Motors) do
      motor.ignitionCoef = 0
    end
    for _, motor in ipairs(p3_Motors) do
      motor.ignitionCoef = 0
    end
    for _, motor in ipairs(motors) do
      motor.ignitionCoef = 0
    end
  end
end

local function updateEHSState(dt)
  --------------------------------------------------------------------------------------------------------------
  ---- this function calculates the aggression in a similar way to game vanilla algorithm, but does not set value to 1 when throttle is greater than 0.7
  local function newAgression(dt)
    local throttle = electrics.values.throttle
  
    if throttle > throttleAggression then
      -- Fast increase in agression
      throttleAggression = throttleAggression + increaseFactor * dt
      if throttleAggression > throttle then
        throttleAggression = throttle
      end
      holdCounter = 0  -- reset hold counter when throttle increases
    elseif throttle < throttleAggression then
      if throttleAggression == 1 and holdCounter < holdTime then
        -- Hold throttleAggression at 1 for a few seconds
        holdCounter = holdCounter + dt
      else
        -- Slow, nonlinear decay in throttleAggression
        local decayRate = max(decayFactor * throttleAggression^2, minDecayFactor)  -- nonlinear decay rate
        throttleAggression = throttleAggression - decayRate * dt
        if throttleAggression < throttle then
            throttleAggression = throttle
        end
      end
    end
    return throttleAggression
  end
  --------------------------------------------------------------------------------------------------------------
  --local vehicleController = require("controller/vehicleController")
  --local aggression = clamp((vehicleController.drivingAggression - 0.2)/0.8, 0, 1)
  --local aggressionCoef = aggression * aggression
  local aggressionCoef = newAgression(dt)

  local enginePower = engine.outputTorque1 * engine.outputAV1
  
  local currentPower = 0
  for _, motor in ipairs(p3_Motors) do
    currentPower = currentPower + (motor.outputTorque1 * motor.outputAV1)
  end
  for _, motor in ipairs(motors) do
    currentPower = currentPower + (motor.outputTorque1 * motor.outputAV1)
  end
  unsmoothedCurrentPower = currentPower
  currentPower = currentPowerSmoother:get(currentPower)
  local currentPowerCoef = currentPowerCoefSmoother:get(currentPower)
  currentPowerCoef = (currentPowerCoef * aggressionCoef + currentPower * (1 - aggressionCoef)) / 1000

  electrics.values.checkRE = engine.isDisabled and 1 or 0
  local soc_related_aggression_threshold = clamp(0.8 - 0.08 * (electrics.values.hybridSOC + 1 - electrics.values.fuel * 100), 0, 100)
  guihooks.message({txt = soc_related_aggression_threshold, time = 0.1})
  -----------------------------------
  -- function for smoothing aggression from linear to exponential, hence reducing RPM at lower agression to improve NVH
  -- parameter aggression should take in correct_aggression, and curvature is optional with default value of 3 (0 = linear, greater number = smoother output)
  local function smoothed_aggression(aggression)
    local maxCurvature = 3
    local curveChangeCoeff = 0.12 --greater value means lower rpm at low battery, better nvh, worse charging capability, suggested value between (0.1 and 0.4)
    local diff = (electrics.values.hybridSOC - electrics.values.fuel)
    local curvature = clamp(curveChangeCoeff / diff, 1, maxCurvature)
    local smoothed = (math.exp(curvature * aggression) - 1) / (math.exp(curvature) - 1)
    return smoothed
  end 
  -----------------------------------

  local function startEngine()
    if (electrics.values.reRunning == 0 and electrics.values.fuel < (electrics.values.hybridSOC - 5) / 100)
      or (aggressionCoef >= min(soc_related_aggression_threshold, 0.99)) 
    then
      if engine.outputAV1 < idleAV * 0.8 then
        engine:deactivateStarter(engine)
        electrics.values.p1RegenThrottle = 0
        electrics.values.p1Throttle = 0.5
      else
        engine:deactivateStarter(engine)
        engine.ignitionCoef = 1
        electrics.values.reRunning = 1
        electrics.values.engineTrottle = throttleSet
        electrics.values.p1Throttle = 0
      end
    end
  end

  local function stopEngine()
    engine:deactivateStarter(engine)
    electrics.values.engineTrottle = 0
    engine.ignitionCoef = 0
    electrics.values.reRunning = 0
    electrics.values.p1Throttle = 0
    electrics.values.p1RegenThrottle = 0
  end

  local function idlingTempKeeper()
    if engine.thermals.coolantTemperature > 90 then
      idlingBreak = 1
    end
    if idlingBreak == 1 and engine.thermals.coolantTemperature < 88 then
      idlingBreak = 0
    end
  end

  ------------------------------------------------------------
  -- User hybrid mode interference logic

  -- set to EV user hybrid mode if engine failure
  if electrics.values.checkRE == 1 then
    electrics.values.hybridMode = "EV"
  end

  -- low battery ev force change to hev mode
  if electrics.values.ignitionLevel == 2 
    and electrics.values.checkRE == 0 
    and electrics.values.fuel < ev_low_battery_exit 
    and electrics.values.hybridMode == "EV" 
  then
    electrics.values.hybridMode = "HEV"
    guihooks.message({txt = "EV mode cancelled due to low battery", time = 3})
  end

  -- high throttle aggression ev force change to hev mode
  if electrics.values.ignitionLevel == 2 
    and electrics.values.checkRE == 0 
    and aggressionCoef > ev_high_throttle_exit 
    and electrics.values.hybridMode == "EV" 
  then
    electrics.values.hybridMode = "HEV"
    guihooks.message({txt = "EV mode temporarily cancelled due to high power request", time = 3})
  end
  ------------------------------------------------------------
  RPMsyncing = false
  if electrics.values.ignitionLevel < 2 or engine.thermals.coolantTemperature >= (engineTempHigh + 10) or electrics.values.hybridMode ~= "HEV" then
    setMode("SERIES")
  elseif electrics.values.ignitionLevel == 2 and electrics.values.hybridMode == "HEV" then
    if electrics.values.fuel >= electrics.values.hybridSOC / 100 or electrics.values.gear ~= "D" then
      setMode("SERIES")
    elseif electrics.values.gear == "D" and electrics.values.fuel < electrics.values.hybridSOC / 100 then
      local p3_OutputAV = 0
      for _, motor in ipairs(p3_Motors) do
        p3_OutputAV = p3_OutputAV + motor.outputAV1
      end
      if #p3_Motors > 0 then
        p3_OutputAV = p3_OutputAV / #p3_Motors
      end

      local engineCanProduceTorque = min((((engine.torqueCurve[floor(p3_OutputAV * iDM_Transmission.gearRatio2 / rpmToAV)] or 0) * engine.intakeAirDensityCoef) * engine.forcedInductionCoef * electrics.values.engineTrottle + engine.nitrousOxideTorque) * engine.outputTorqueState * engine.slowIgnitionErrorCoef * engine.fastIgnitionErrorCoef, engine.maxTorqueLimit)

      if electrics.values.reRunning ~= 1 
        or electrics.values.brake > 0 
        or p3_OutputAV > freeAV / iDM_Transmission.gearRatio2 then
        setMode("SERIES")
      elseif electrics.values.reRunning == 1 and electrics.values.brake == 0 and p3_OutputAV <= freeAV / iDM_Transmission.gearRatio2 then
        if electrics.values.fuel >= batteryCriticalLow then
          if (p3_OutputAV < parallelAV / iDM_Transmission.gearRatio2 * parallel_to_series) 
          -- or ((engineCanProduceTorque * p3_OutputAV * iDM_Transmission.gearRatio2 * 0.975 < currentPower) and (enginePower < currentPower) and (aggressionCoef <= 0.25)) 
            or (aggressionCoef > soc_related_aggression_threshold * parallel_to_series_threshold_factor) then
            setMode("SERIES")
          elseif (p3_OutputAV > parallelAV / iDM_Transmission.gearRatio2 * series_to_parallel) 
           -- and ((engineCanProduceTorque * p3_OutputAV * iDM_Transmission.gearRatio2 > currentPower) or (aggressionCoef > 0.35)) 
           -- and (aggressionCoef < 0.4)
          then
              if abs(engine.outputAV1 - p3_OutputAV * iDM_Transmission.gearRatio2) > 10 then
                RPMsyncing = true
              else
                RPMsyncing = false
                setMode("PARALLEL")             
              end
          end
        else
          if (p3_OutputAV < parallelAV / iDM_Transmission.gearRatio2 * parallel_to_series) or (aggressionCoef > 0.2) then
            setMode("SERIES")
          elseif (p3_OutputAV > parallelAV / iDM_Transmission.gearRatio2 * series_to_parallel) and (aggressionCoef <= 0.35) then
            if abs(engine.outputAV1 - p3_OutputAV * iDM_Transmission.gearRatio2) > 10 then
              RPMsyncing = true
            else
              RPMsyncing = false
              setMode("PARALLEL")             
            end
          end
        end
      end
    end
  end

  if electrics.values.hevState == "PARALLEL" then
    engine:setTempRevLimiter(freeAV)
    startEngine()
    electrics.values.p1Throttle = 0
    electrics.values.p1RegenThrottle = 0
    engine.electricsThrottleName = "engineTrottle"
    electrics.values.engineTrottle = min(throttleSet + (electrics.values.engineThrottleOffset or 0), 1)
    ------------------------------------------------------------
    -- engine off when off gas for parallel mode
    if electrics.values.fuel > ev_low_battery_exit 
      and aggressionCoef < min(soc_related_aggression_threshold * parallel_to_ev_threshold_factor, parallel_to_ev_threshold_factor)
      and electrics.values.wheelspeed * 3.6 <= max_engine_stop_speed
      and electrics.values.hevState == "PARALLEL"
    then
      electrics.values.hevState = "SERIES"
      stopEngine()
    end
    ------------------------------------------------------------
  else --series
    local p3_OutputAV = 0
      for _, motor in ipairs(p3_Motors) do
        p3_OutputAV = p3_OutputAV + motor.outputAV1
      end
      if #p3_Motors > 0 then
        p3_OutputAV = p3_OutputAV / #p3_Motors
      end
    engine.electricsThrottleName = "engineTrottle"
    local speedAV = p3_OutputAV * iDM_Transmission.gearRatio2

    if electrics.values.ignitionLevel < 2 or electrics.values.hybridMode == "EV"  or engine.thermals.coolantTemperature >= (engineTempHigh + 10) then
      idlingTime = 0
      stopEngine()
    elseif electrics.values.ignitionLevel == 2 then
      idlingTempKeeper()
      if electrics.values.wheelspeed <= idleSpeed and electrics.values.fuel >= batteryLow and aggressionCoef < 0.75 and idlingBreak == 1 then
        engine:setTempRevLimiter(idleAV)
        idlingTime = idlingTime + dt
        if idlingTime >= 15 then
          stopEngine()
        end
       ------------------------------------------------------------
      -- engine off when off gas for series mode
      elseif electrics.values.fuel > ev_low_battery_exit 
        and aggressionCoef <= min(soc_related_aggression_threshold * series_to_ev_threshold_factor, series_to_ev_threshold_factor)
        and electrics.values.hevState == "SERIES"
      then
        stopEngine()
    ------------------------------------------------------------
      elseif engine.thermals.coolantTemperature > engineTempHigh then
        idlingTime = 0
        engine:setTempRevLimiter(idleAV)
        if engine.thermals.coolantTemperature < (engineTempHigh + 5) then
          startEngine()
        end
      elseif electrics.values.fuel >= batteryCriticalLow then
        idlingTime = 0
        -- local minRPM = clamp(currentPowerCoef * 100, idleRPM, freeRPM)
        -- local maxRPM = clamp(sqrt(electrics.values.wheelspeed) * 800, minRPM, freeRPM)
        -- engine:setTempRevLimiter((minRPM + (maxRPM - minRPM) * aggressionCoef) * rpmToAV)
        if RPMsyncing == false then
          engine:setTempRevLimiter(max(idleAV, smoothed_aggression(aggressionCoef) * (redlineRPM * rpmToAV - speedAV) + speedAV + 200 * rpmToAV))
        elseif RPMsyncing then
          engine:setTempRevLimiter(speedAV)
        end
        startEngine()
      else
        idlingTime = 0
        -- local minRPM = clamp(currentPowerCoef * 120, idleRPM, freeRPM)
        -- local maxRPM = clamp(sqrt(electrics.values.wheelspeed) * 1200, idleRPM, freeRPM)
        -- local maxRPM = freeRPM
        -- engine:setTempRevLimiter((minRPM + (maxRPM - minRPM) * aggressionCoef) * rpmToAV)
        if RPMsyncing == false then
          engine:setTempRevLimiter(max(idleAV, smoothed_aggression(aggressionCoef) * (redlineRPM * rpmToAV - speedAV) + speedAV + 500 * rpmToAV))
        elseif RPMsyncing then
          engine:setTempRevLimiter(speedAV)
        end
        startEngine()
      end
    end

    if electrics.values.reRunning == 1 then
      electrics.values.engineTrottle = throttleSet
      if engine.outputAV1 < engine.tempRevLimiterAV * 0.9 then
        electrics.values.p1RegenThrottle = max(0, electrics.values.p1RegenThrottle - 0.02)
      end
      if engine.throttle < throttleSet then
        electrics.values.p1RegenThrottle = min(1, electrics.values.p1RegenThrottle + 0.02)
      end
    else
      electrics.values.engineTrottle = 0
    end
  end
end

local function updateGaugeInfo(dt)
  local totalCapacity = 0
  local remainingEnergy = 0
  local remainingVolume = 0
  
  for _, s in pairs(engine.registeredEnergyStorages) do
    local storage = energyStorage.getStorage(s)
    if storage and storage.type ~= "n2oTank" then
      totalCapacity = totalCapacity + storage.energyCapacity
      remainingEnergy = remainingEnergy + storage.storedEnergy
      remainingVolume = remainingVolume + storage.remainingVolume
    end
  end
  local remainingGasolineRatio = remainingEnergy / totalCapacity
  electrics.values.remainingGasolineRatio = remainingGasolineRatio
  if remainingGasolineRatio <= 0.1 then
    electrics.values.lowGasoline = 1
  else
    electrics.values.lowGasoline = 0
  end

  local motorThrottle = electrics.values.throttle
  local motorRegenThrottle = electrics.values.regenThrottle
  local motorCount = #p3_Motors + #motors
  electrics.values.motorCount = motorCount

  if electrics.values.hevState == "PARALLEL" then
    if electrics.values.throttle > 0 then
      electrics.values.hybridEnergyFlow = "P-"
    elseif electrics.values.throttle <= 0 and electrics.values.regenThrottle == 0 then
      electrics.values.hybridEnergyFlow = "PP"
    elseif electrics.values.throttle <= 0 and electrics.values.regenThrottle > 0 then
      if unsmoothedCurrentPower <= 0 then
        electrics.values.hybridEnergyFlow = "PR"
      else
        electrics.values.hybridEnergyFlow = "P+"
      end
    end
  elseif electrics.values.hevState == "SERIES" then
    if electrics.values.reRunning == 1 and electrics.values.wheelspeed >= 0.1 then
      if motorThrottle > 0 then
        if electrics.values.fuel < lastBatteryLevel then
          electrics.values.hybridEnergyFlow = "S-"
        elseif electrics.values.fuel == lastBatteryLevel then
          electrics.values.hybridEnergyFlow = "SS"
        elseif electrics.values.fuel > lastBatteryLevel then
          electrics.values.hybridEnergyFlow = "S+"
        end
      elseif motorThrottle <= 0 and motorRegenThrottle == 0 then
        electrics.values.hybridEnergyFlow = "SC"
      elseif motorThrottle <= 0 and motorRegenThrottle > 0 then
        electrics.values.hybridEnergyFlow = "SR"
      end
    elseif electrics.values.reRunning == 0 and electrics.values.wheelspeed >= 0.1 then
      if motorThrottle > 0 then
        electrics.values.hybridEnergyFlow = "E-"
      elseif motorThrottle <= 0 and motorRegenThrottle == 0 then
        electrics.values.hybridEnergyFlow = "EE"
      elseif motorThrottle <= 0 and motorRegenThrottle > 0 then
        electrics.values.hybridEnergyFlow = "ER"
      end
    elseif electrics.values.wheelspeed < 0.1 then
      if electrics.values.reRunning == 1 then
        electrics.values.hybridEnergyFlow = "SC"
      else
        electrics.values.hybridEnergyFlow = "EE"
      end
    end
  end
  lastBatteryLevel = electrics.values.fuel
end



local function updateGFX(dt)
  updateMotorIgnition()
  updateEHSState(dt)
  updateGaugeInfo(dt)
end

local function serialize()
end

local function deserialize(data)
end

local function init(jbeamData)
  extensions.load("socControl")
  engine = powertrain.getDevice(jbeamData.engineName) or powertrain.getDevice("mainEngine")
  iDM_Transmission = powertrain.getDevice(jbeamData.iDM_Transmission_Name) or powertrain.getDevice(jbeamData.p1Name)
  p1_Motors = powertrain.getDevicesByType("p1_Motor")
  p3_Motors = powertrain.getDevicesByType("p3_Motor")
  motors = powertrain.getDevicesByType("electricMotor")
  batteryLow = jbeamData.batteryLow or 0.15
  batteryCriticalLow = jbeamData.batteryCriticalLow or 0.05
  freeRPM = jbeamData.freeRPM or 5500
  idleRPM = jbeamData.idleRPM or 1500
  parallelRPM = jbeamData.parallelRPM or 1200
  freeAV = freeRPM * rpmToAV
  idleAV = idleRPM * rpmToAV
  parallelAV = parallelRPM * rpmToAV
  idleSpeed = (jbeamData.idleSpeed or 15) * kphToMs

  electrics.values.checkRE = engine.isDisabled and 1 or 0

  local minV = nil
  for k, v in pairs(engine.invBurnEfficiencyTable) do
    if minV == nil or v < minV or (v == minV and k  < throttleSet) then
      minV = v
      throttleSet = k
    end
  end
  throttleSet = throttleSet / 100
end

local function initLastStage(jbeamData)
end

M.init = init
M.initLastStage = initLastStage
M.updateMotorIgnition = updateMotorIgnition
M.updateEHSState = updateEHSState
M.updateGFX = updateGFX
M.serialize = serialize
M.deserialize = deserialize

return M
